import uuid

from django.urls import reverse
from rest_framework import status

from arkindex.dataimport.models import DataImport, DataImportMode, WorkerVersion
from arkindex.documents.models import Classification, ClassificationState, Corpus, Element, MLClass
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import User


class TestProcessElements(FixtureAPITestCase):
    """
    Test list process elements
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        # ----------------- TREE ------------------
        #
        #        Folder 1          Folder 2
        #          /|\               /|\
        #         / | \             / | \
        #        /  |  \           /  |  \
        #       /   |   \         /   |   \
        #      /    |    \       /    |    \
        #     /     |     \     /     |     \
        #    /      |      \   /      |      \
        #   /       |       \ /       |       \
        # Page 1  Page 2   Page 3  Page 4  Page 5
        #   |      |         |       |       |
        # Line 1  Line 2   Line 3  Line 4  Line 5

        # Create a corpus
        cls.private_corpus = Corpus.objects.create(name='private')

        # Create element types
        cls.folder_type, _ = cls.private_corpus.types.get_or_create(
            slug='folder',
            display_name='Folder',
            folder=True,
        )
        cls.page_type, _ = cls.private_corpus.types.get_or_create(
            slug='page',
            display_name='Page',
        )
        cls.line_type, _ = cls.private_corpus.types.get_or_create(
            slug='text_line',
            display_name='Line',
        )

        # Create 2 volumes
        cls.folder_1 = Element.objects.create(
            corpus=cls.private_corpus,
            type=cls.folder_type,
            name="Baba au rhum",
        )
        cls.folder_2 = Element.objects.create(
            corpus=cls.private_corpus,
            type=cls.folder_type,
            name="Mongolfiere",
        )

        # Create 2 pages in each volume and 1 common page
        cls.page_1 = Element.objects.create(
            corpus=cls.private_corpus,
            name="Baba au rhum 1",
            type=cls.page_type
        )
        cls.page_2 = Element.objects.create(
            corpus=cls.private_corpus,
            name="Mongolfiere 1",
            type=cls.page_type
        )
        cls.page_3 = Element.objects.create(
            corpus=cls.private_corpus,
            name="Common",
            type=cls.page_type
        )
        cls.page_4 = Element.objects.create(
            corpus=cls.private_corpus,
            name="Mongolfiere 2",
            type=cls.page_type
        )
        cls.page_5 = Element.objects.create(
            corpus=cls.private_corpus,
            name="Baba au rhum 2",
            type=cls.page_type
        )
        cls.page_1.add_parent(cls.folder_1)
        cls.page_2.add_parent(cls.folder_1)
        cls.page_3.add_parent(cls.folder_1)
        cls.page_4.add_parent(cls.folder_2)
        cls.page_5.add_parent(cls.folder_2)
        cls.page_3.add_parent(cls.folder_2)

        # Create 1 lines in each page
        cls.line_1 = Element.objects.create(
            corpus=cls.private_corpus,
            name="Line 1",
            type=cls.line_type
        )
        cls.line_2 = Element.objects.create(
            corpus=cls.private_corpus,
            name="Line 2",
            type=cls.line_type
        )
        cls.line_3 = Element.objects.create(
            corpus=cls.private_corpus,
            name="Line 3",
            type=cls.line_type
        )
        cls.line_4 = Element.objects.create(
            corpus=cls.private_corpus,
            name="Line 4",
            type=cls.line_type
        )
        cls.line_5 = Element.objects.create(
            corpus=cls.private_corpus,
            name="Line 5",
            type=cls.line_type
        )
        cls.line_1.add_parent(cls.page_1)
        cls.line_2.add_parent(cls.page_2)
        cls.line_3.add_parent(cls.page_3)
        cls.line_4.add_parent(cls.page_4)
        cls.line_5.add_parent(cls.page_5)

        # Create best classes
        worker_version = WorkerVersion.objects.get(worker__slug='reco')
        cls.coffee_class = MLClass.objects.create(name='C0FFEE', corpus=cls.private_corpus)
        cls.food_class = MLClass.objects.create(name='F00D', corpus=cls.private_corpus)
        Classification.objects.create(
            element=cls.folder_2,
            state=ClassificationState.Validated,
            ml_class=cls.food_class,
            worker_version=worker_version,
        )
        Classification.objects.create(
            element=cls.page_1,
            state=ClassificationState.Validated,
            ml_class=cls.coffee_class,
            worker_version=worker_version,
        )
        Classification.objects.create(
            element=cls.page_2,
            high_confidence=True,
            ml_class=cls.food_class,
            worker_version=worker_version,
        )
        Classification.objects.create(
            element=cls.page_3,
            state=ClassificationState.Validated,
            ml_class=cls.food_class,
            worker_version=worker_version,
        )
        Classification.objects.create(
            element=cls.page_5,
            high_confidence=True,
            ml_class=cls.food_class,
            worker_version=worker_version,
        )
        Classification.objects.create(
            element=cls.page_5,
            state=ClassificationState.Validated,
            ml_class=cls.coffee_class,
            worker_version=worker_version,
        )

    def setUp(self):
        super().setUp()
        self.dataimport = DataImport.objects.create(
            creator_id=self.user.id,
            mode=DataImportMode.Workers,
            corpus_id=self.private_corpus.id
        )

    def test_requires_login(self):
        response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_wrong_id(self):
        self.client.force_login(self.superuser)
        response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': str(uuid.uuid4())}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_no_access(self):
        self.dataimport.corpus = Corpus.objects.create(name='private')
        self.corpus.creator = User.objects.create_user('John Doe')
        self.dataimport.save()
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have an admin access to the corpus of this process.'})

    def test_filter_elements_wrong_corpus(self):
        """
        Selected elements must be part of the same corpus as the process
        """
        self.dataimport.corpus = Corpus.objects.create(name='private')
        self.dataimport.save()
        self.dataimport.elements.add(self.page_1.id, self.folder_2.id)

        self.client.force_login(self.superuser)
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': self.dataimport.id}))
        self.assertDictEqual(response.json(), {
            '__all__': [f'Some elements on this process are not part of corpus {self.dataimport.corpus_id}']
        })

    def test_filter_elements_multiple_corpus(self):
        """
        Selected elements are part of multiple corpora
        This should not happen in usual situations
        """
        corpus = Corpus.objects.create(name='private')
        element = Element.objects.create(
            corpus=corpus,
            type=self.folder_type,
            name="Folder",
        )
        self.dataimport.elements.add(element.id, self.page_1.id)

        self.client.force_login(self.superuser)
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': self.dataimport.id}))
        self.assertDictEqual(response.json(), {
            '__all__': [f'Some elements on this process are not part of corpus {self.dataimport.corpus_id}']
        })

    def test_filter_element_wrong_corpus(self):
        """
        Element and corpus are different
        This should not happen in usual situations
        """
        self.dataimport.corpus = Corpus.objects.create(name='private')
        self.dataimport.element = self.page_1
        self.dataimport.save()

        self.client.force_login(self.superuser)
        with self.assertNumQueries(5):
            response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': self.dataimport.id}))
        self.assertDictEqual(response.json(), {
            '__all__': [f'Element {self.page_1.id} is not part of corpus {self.dataimport.corpus_id}']
        })

    def test_filter_name(self):
        self.dataimport.name_contains = "rhum"
        self.dataimport.save()
        elements = [self.folder_1, self.page_1, self.page_5]

        self.client.force_login(self.superuser)
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], None)
        self.assertEqual(data["next"], None)
        self.assertCountEqual(data["results"], [
            {
                'id': str(element.id),
                'type': element.type.slug,
                'name': element.name
            }
            for element in elements
        ])

    def test_filter_name_from_element(self):
        self.dataimport.element = self.folder_1
        self.dataimport.name_contains = "rhum"
        self.dataimport.load_children = True
        self.dataimport.save()
        elements = [self.folder_1, self.page_1]

        self.client.force_login(self.superuser)
        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], None)
        self.assertEqual(data["next"], None)
        self.assertCountEqual(data["results"], [
            {
                'id': str(element.id),
                'type': element.type.slug,
                'name': element.name
            }
            for element in elements
        ])

    def test_filter_type(self):
        self.dataimport.element_type = self.folder_type
        self.dataimport.save()
        elements = [self.folder_1, self.folder_2]

        self.client.force_login(self.superuser)
        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], None)
        self.assertEqual(data["next"], None)
        self.assertCountEqual(data["results"], [
            {
                'id': str(element.id),
                'type': element.type.slug,
                'name': element.name
            }
            for element in elements
        ])

    def test_filter_type_from_element(self):
        self.dataimport.element = self.folder_1
        self.dataimport.element_type = self.line_type
        self.dataimport.load_children = True
        self.dataimport.save()
        elements = [self.line_1, self.line_2, self.line_3]

        self.client.force_login(self.superuser)
        with self.assertNumQueries(8):
            response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], None)
        self.assertEqual(data["next"], None)
        self.assertCountEqual(data["results"], [
            {
                'id': str(element.id),
                'type': element.type.slug,
                'name': element.name
            }
            for element in elements
        ])

    def test_filter_best_class_by_id(self):
        self.dataimport.best_class = self.food_class.id
        self.dataimport.save()
        elements = [self.page_5, self.page_3, self.folder_2, self.page_2]

        self.client.force_login(self.superuser)
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], None)
        self.assertEqual(data["next"], None)
        self.assertCountEqual(data["results"], [
            {
                'id': str(element.id),
                'type': element.type.slug,
                'name': element.name
            }
            for element in elements
        ])

    def test_filter_any_best_class(self):
        self.dataimport.best_class = "true"
        self.dataimport.save()
        elements = [self.page_1, self.page_5, self.page_3, self.folder_2, self.page_2]

        self.client.force_login(self.superuser)
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], None)
        self.assertEqual(data["next"], None)
        self.assertCountEqual(data["results"], [
            {
                'id': str(element.id),
                'type': element.type.slug,
                'name': element.name
            }
            for element in elements
        ])

    def test_filter_no_best_class(self):
        self.dataimport.best_class = "false"
        self.dataimport.save()
        elements = [self.folder_1, self.line_1, self.line_2, self.line_3, self.line_4, self.line_5, self.page_4]

        self.client.force_login(self.superuser)
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], None)
        self.assertEqual(data["next"], None)
        self.assertCountEqual(data["results"], [
            {
                'id': str(element.id),
                'type': element.type.slug,
                'name': element.name
            }
            for element in elements
        ])

    def test_filter_element(self):
        self.dataimport.element = self.page_1
        self.dataimport.save()

        self.client.force_login(self.superuser)
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], None)
        self.assertEqual(data["next"], None)
        self.assertCountEqual(data["results"], [
            {
                'id': str(self.page_1.id),
                'type': self.page_1.type.slug,
                'name': self.page_1.name
            }
        ])

    def test_filter_elements(self):
        self.dataimport.elements.add(self.page_1.id, self.folder_2.id)

        self.client.force_login(self.superuser)
        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], None)
        self.assertEqual(data["next"], None)
        self.assertCountEqual(data["results"], [
            {
                'id': str(self.page_1.id),
                'type': self.page_1.type.slug,
                'name': self.page_1.name
            },
            {
                'id': str(self.folder_2.id),
                'type': self.folder_2.type.slug,
                'name': self.folder_2.name
            }
        ])

    def test_load_children_and_filter_name(self):
        self.dataimport.name_contains = "rhum"
        self.dataimport.load_children = True
        self.dataimport.save()
        elements = [self.folder_1, self.page_1, self.page_5]

        self.client.force_login(self.superuser)
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], None)
        self.assertEqual(data["next"], None)
        self.assertCountEqual(data["results"], [
            {
                'id': str(element.id),
                'type': element.type.slug,
                'name': element.name
            }
            for element in elements
        ])

    def test_corpus_and_filter_type(self):
        self.dataimport.element_type = self.folder_type
        self.dataimport.load_children = True
        self.dataimport.save()
        elements = [self.folder_1, self.folder_2]

        self.client.force_login(self.superuser)
        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], None)
        self.assertEqual(data["next"], None)
        self.assertCountEqual(data["results"], [
            {
                'id': str(element.id),
                'type': element.type.slug,
                'name': element.name
            }
            for element in elements
        ])

    def test_load_children_and_filter_best_class_by_id(self):
        self.dataimport.best_class = self.food_class.id
        self.dataimport.load_children = True
        self.dataimport.save()
        elements = [self.folder_2, self.page_2, self.page_3, self.page_5]

        self.client.force_login(self.superuser)
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], None)
        self.assertEqual(data["next"], None)
        self.assertCountEqual(data["results"], [
            {
                'id': str(element.id),
                'type': element.type.slug,
                'name': element.name
            }
            for element in elements
        ])

    def test_load_children_and_filter_best_class(self):
        self.dataimport.best_class = "true",
        self.dataimport.load_children = True
        self.dataimport.save()
        elements = [self.page_1, self.page_5, self.page_3, self.folder_2, self.page_2]

        self.client.force_login(self.superuser)
        with self.assertNumQueries(6):
            response = self.client.get(
                reverse('api:process-elements-list', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], None)
        self.assertEqual(data["next"], None)
        self.assertCountEqual(data["results"], [
            {
                'id': str(element.id),
                'type': element.type.slug,
                'name': element.name
            }
            for element in elements
        ])

    def test_load_children_and_filter_no_best_class(self):
        self.dataimport.best_class = "false"
        self.dataimport.load_children = True
        self.dataimport.save()
        elements = [self.folder_1, self.line_1, self.line_2, self.line_3, self.line_4, self.line_5, self.page_4]

        self.client.force_login(self.superuser)
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], None)
        self.assertEqual(data["next"], None)
        self.assertCountEqual(data["results"], [
            {
                'id': str(element.id),
                'type': element.type.slug,
                'name': element.name
            }
            for element in elements
        ])

    def test_load_children_and_filter_element(self):
        self.dataimport.element = self.folder_1
        self.dataimport.load_children = True
        self.dataimport.save()
        elements = [self.folder_1, self.page_1, self.page_3, self.line_1, self.line_2, self.line_3, self.page_2]

        self.client.force_login(self.superuser)
        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], None)
        self.assertEqual(data["next"], None)
        self.assertCountEqual(data["results"], [
            {
                'id': str(element.id),
                'type': element.type.slug,
                'name': element.name
            }
            for element in elements
        ])

    def test_load_children_and_filter_elements(self):
        self.dataimport.load_children = True
        self.dataimport.save()
        self.dataimport.elements.add(self.page_1.id, self.folder_2.id)
        elements = [self.page_1, self.page_5, self.page_3, self.line_1, self.line_3, self.line_4, self.line_5, self.folder_2, self.page_4]

        self.client.force_login(self.superuser)
        with self.assertNumQueries(8):
            response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], None)
        self.assertEqual(data["next"], None)
        self.assertCountEqual(data["results"], [
            {
                'id': str(element.id),
                'type': element.type.slug,
                'name': element.name
            }
            for element in elements
        ])

    def test_load_children_and_any_filter(self):
        self.dataimport.load_children = True
        self.dataimport.save()
        elements = Element.objects.filter(corpus=self.private_corpus).order_by('name', 'type__slug')

        self.client.force_login(self.superuser)
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data["count"], None)
        self.assertEqual(data["next"], None)
        self.assertCountEqual(data["results"], [
            {
                'id': str(element.id),
                'type': element.type.slug,
                'name': element.name
            }
            for element in elements
        ])

    def test_all_modes(self):
        self.dataimport.load_children = True
        self.dataimport.save()
        self.client.force_login(self.superuser)

        # Every mode other than Workers never returns elements
        for mode in (DataImportMode.Images, DataImportMode.PDF, DataImportMode.Repository, DataImportMode.IIIF, DataImportMode.Transkribus):
            self.dataimport.mode = mode
            self.dataimport.save()
            with self.assertNumQueries(4):
                response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': self.dataimport.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)
            data = response.json()
            self.assertCountEqual(data, {'count': None, 'next': None, 'previous': None, 'results': []})

        elements = Element.objects.filter(corpus=self.private_corpus).order_by('name', 'type__slug')

        # The Workers mode returns some elements
        self.dataimport.mode = DataImportMode.Workers
        self.dataimport.save()
        with self.assertNumQueries(6):
            response = self.client.get(reverse('api:process-elements-list', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertIsNone(data["count"])
        self.assertIsNone(data["next"])
        self.assertCountEqual(data["results"], [
            {
                'id': str(element.id),
                'type': element.type.slug,
                'name': element.name
            }
            for element in elements
        ])

    def test_list_elements_cursor_pagination(self):
        """
        No element is duplicated or dropped
        """
        elts = Element.objects.bulk_create([
            Element(
                corpus=self.private_corpus,
                name='Similar name',
                type=self.page_type
            ) for i in range(40)
        ])
        for elt in elts:
            elt.add_parent(self.folder_1)
        self.dataimport.element = self.folder_1
        self.dataimport.name_contains = 'Similar'
        self.dataimport.load_children = True
        self.dataimport.save()

        self.client.force_login(self.superuser)
        with self.assertNumQueries(7):
            page_1 = self.client.get(reverse('api:process-elements-list', kwargs={'pk': self.dataimport.id}))
        self.assertEqual(len(page_1.json()['results']), 20)
        next_page = page_1.json().get('next')
        self.assertIsNotNone(next_page)
        with self.assertNumQueries(7):
            page_2 = self.client.get(next_page)
        self.assertIsNone(page_2.json()['next'])
        qs_1 = Element.objects.filter(id__in=[elt['id'] for elt in page_1.json()['results']])
        qs_2 = Element.objects.filter(id__in=[elt['id'] for elt in page_2.json()['results']])

        self.assertEqual(qs_1.intersection(qs_2).count(), 0)
        self.assertEqual(qs_1.union(qs_2).distinct().count(), 40)

    def test_elements_invalid_cursor(self):
        self.client.force_login(self.superuser)
        response = self.client.get(
            reverse('api:process-elements-list', kwargs={'pk': self.dataimport.id}),
            {'cursor': 'ABC', 'with_count': True}
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(response.json(), {'detail': 'Invalid cursor'})

    def test_elements_count(self):
        """
        Elements count can be retrieved with with_count parameter
        """
        self.client.force_login(self.superuser)
        with self.assertNumQueries(7):
            response = self.client.get(
                reverse('api:process-elements-list', kwargs={'pk': self.dataimport.id}),
                {'page_size': 6, 'with_count': True}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertIsNotNone(data['next'])
        self.assertEqual(data['count'], 12)
        self.assertEqual(len(data['results']), 6)
        next_url = data.get('next')
        self.assertIn('cursor', next_url)

        second_page = self.client.get(next_url)
        data = second_page.json()
        # Count should be present in the next URL
        self.assertEqual(data['count'], 12)
        self.assertIsNone(data['next'])
        self.assertEqual(len(data['results']), 6)

    def test_cursor_pagination_page_size(self):
        """
        Page size may be changed for cursor pagination
        """
        Element.objects.bulk_create([
            Element(
                corpus=self.private_corpus,
                name='Similar name',
                type=self.page_type
            ) for i in range(51)
        ])
        self.dataimport.name_contains = 'Similar'
        self.dataimport.save()

        self.client.force_login(self.superuser)
        with self.assertNumQueries(6):
            response = self.client.get(
                reverse('api:process-elements-list', kwargs={'pk': self.dataimport.id}),
                {'page_size': 50}
            )
        data = response.json()
        self.assertEqual(len(data['results']), 50)
        self.assertIsNotNone(data['next'])
