from unittest.mock import patch

from django.urls import reverse
from rest_framework import status

from arkindex.dataimport.models import RepositoryType
from arkindex.dataimport.providers import GitLabProvider, GitProvider
from arkindex.project.tests import FixtureAPITestCase


class TestProviders(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.creds = cls.user.credentials.get()
        cls.repo = cls.creds.repos.get(type=RepositoryType.IIIF)
        cls.rev = cls.repo.revisions.get()

    def test_init(self):
        glp = GitLabProvider()
        self.assertEqual(glp.url, GitLabProvider.url)

        glp = GitLabProvider(url='something')
        self.assertEqual(glp.url, 'something')

        glp = GitLabProvider(credentials=self.creds)
        self.assertEqual(glp.url, self.creds.provider_url)

        with self.assertRaises(Exception):
            GitLabProvider(credentials='not a OAuthCredentials')

        glp = GitLabProvider(credentials=self.creds, url='something')
        self.assertEqual(glp.url, 'something')

    @patch('arkindex.dataimport.api.Repository.provider_class')
    def test_webhook_no_credentials(self, provider_class):
        self.client.force_login(self.user)
        self.repo.credentials = None
        self.repo.save()
        response = self.client.post(reverse('api:import-hook', kwargs={'pk': self.repo.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'No credentials available for this repository.'})

    @patch('arkindex.dataimport.api.Repository.provider_class')
    def test_webhook(self, provider_class):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-hook', kwargs={'pk': self.repo.id}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertTrue(provider_class.return_value.handle_webhook.called)

    @patch('arkindex.dataimport.models.DataImport.start')
    def test_start_imports_workers(self, start_di_mock):
        self.repo.type = RepositoryType.Worker
        self.repo.save()
        self.assertEqual(self.rev.dataimports.count(), 0)
        GitProvider.start_imports(None, rev=self.rev)
        self.assertEqual(self.rev.dataimports.count(), 1)
        self.assertEqual(start_di_mock.call_count, 1)
        _, start_kwargs = start_di_mock.call_args
        self.assertDictEqual(start_kwargs, {'thumbnails': False})

    @patch('arkindex.dataimport.models.DataImport.start')
    def test_start_imports_iiif(self, start_di_mock):
        corpora = [self.repo.corpora.create(), self.corpus]
        self.assertEqual(self.rev.dataimports.count(), 0)
        GitProvider.start_imports(None, rev=self.rev)
        self.assertEqual(self.rev.dataimports.count(), 2)
        self.assertEqual(start_di_mock.call_count, 2)
        self.assertCountEqual(
            start_di_mock.call_args_list,
            [((), {'thumbnails': True, 'corpus_id': c.id}) for c in corpora]
        )
