from collections import defaultdict
from uuid import UUID

from django.contrib.gis.geos import LinearRing
from django.db import transaction
from django.utils.functional import cached_property
from drf_spectacular.utils import extend_schema_field
from rest_framework import serializers
from rest_framework.exceptions import ValidationError

from arkindex.dataimport.models import WorkerVersion
from arkindex.documents.models import (
    Corpus,
    Element,
    ElementPath,
    ElementType,
    Entity,
    MetaData,
    Transcription,
    TranscriptionEntity,
)
from arkindex.documents.serializers.entities import BaseEntitySerializer, TranscriptionEntityDetailsSerializer
from arkindex.documents.serializers.light import (
    CorpusLightSerializer,
    ElementLightSerializer,
    ElementTypeLightSerializer,
    MetaDataLightSerializer,
)
from arkindex.documents.serializers.ml import ClassificationSerializer
from arkindex.images.models import Image, Zone
from arkindex.images.serializers import ZoneSerializer
from arkindex.project.serializer_fields import LinearRingField
from arkindex.users.models import Role
from arkindex.users.utils import get_max_level


class MetaDataSerializer(MetaDataLightSerializer):
    """
    Serialises some Metadata for any Element
    """
    entity = BaseEntitySerializer()

    class Meta:
        model = MetaData
        fields = MetaDataLightSerializer.Meta.fields + ('entity', )


class MetaDataUpdateSerializer(MetaDataLightSerializer):
    """
    Allow editing MetaData
    """
    entity = serializers.PrimaryKeyRelatedField(queryset=Entity.objects.none(), required=False, allow_null=True)
    worker_version = serializers.PrimaryKeyRelatedField(queryset=WorkerVersion.objects.all(), required=False, allow_null=True)

    class Meta:
        model = MetaData
        fields = MetaDataLightSerializer.Meta.fields + ('entity', 'worker_version', )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # We don't want to return the entity field when listing metadata
        if self.context.get('list'):
            self.fields.pop('entity')
            return

        if self.context.get('element'):
            corpus = self.context['element'].corpus
        elif self.instance:
            corpus = self.instance.element.corpus
        else:
            return
        self.fields['entity'].queryset = corpus.entities.all()


class CorpusSerializer(serializers.ModelSerializer):
    """
    Serialize a corpus
    """
    rights = serializers.SerializerMethodField(read_only=True)
    types = ElementTypeLightSerializer(many=True, read_only=True)
    authorized_users = serializers.SerializerMethodField(read_only=True)

    class Meta:
        model = Corpus
        fields = (
            'id',
            'name',
            'description',
            'public',
            'rights',
            'types',
            'created',
            'authorized_users',
        )

    @extend_schema_field(serializers.ListField(serializers.CharField()))
    def get_rights(self, corpus):

        if getattr(corpus, 'access_level', None):
            # Use annotated rights if available
            level = corpus.access_level
        else:
            level = get_max_level(self.context['request'].user, corpus)

        rights = ['read']
        if level >= Role.Contributor.value:
            rights.append('write')
        if level >= Role.Admin.value:
            rights.append('admin')
        return rights

    def get_authorized_users(self, corpus) -> int:
        count = getattr(corpus, 'authorized_users', None)
        if count is None:
            count = corpus.memberships.count()
        return count

    def validate_public(self, public):
        """
        Only an admin can toggle a corpus public property
        Normal users may create private corpus only
        """
        toggled = self.instance and self.instance.public != public
        if (toggled or public) and not self.context['request'].user.is_admin:
            raise ValidationError(['Only admin users are allowed to edit the public attribute on a corpus.'])
        return public

    @transaction.atomic
    def create(self, validated_data):
        corpus = Corpus.objects.create(**validated_data)
        corpus.memberships.create(
            user=self.context['request'].user,
            level=Role.Admin.value
        )
        corpus.create_default_types()
        # Cheat annotating corpus to avoid a stale read when serializing it
        corpus.access_level = Role.Admin.value
        corpus.authorized_users = 1
        return corpus


class ElementTinySerializer(serializers.ModelSerializer):
    """
    Serialises a document
    """
    type = serializers.SlugRelatedField(queryset=ElementType.objects.none(), slug_field='slug')
    corpus = CorpusLightSerializer(read_only=True)
    thumbnail_url = serializers.SerializerMethodField(read_only=True)
    zone = ZoneSerializer(read_only=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('element'):
            # Do not raise Error in order to create OpenAPI schema
            return
        self.fields['type'].queryset = ElementType.objects.filter(corpus=self.context['element'].corpus)

    @extend_schema_field(serializers.CharField(allow_null=True))
    def get_thumbnail_url(self, element):
        if element.type.folder:
            return element.thumbnail.s3_url

    class Meta:
        model = Element
        fields = (
            'id',
            'type',
            'name',
            'corpus',
            'thumbnail_url',
            'zone',
        )
        read_only_fields = (
            'id',
            'corpus',
            'thumbnail_url',
            'zone',
        )


class ElementSlimSerializer(ElementTinySerializer):
    """
    Fully serialises a document
    """
    thumbnail_put_url = serializers.SerializerMethodField(read_only=True)

    @extend_schema_field(serializers.CharField(allow_null=True))
    def get_thumbnail_put_url(self, element):
        user = self.context['request'].user
        if user.is_authenticated and (user.is_admin or user.is_internal) and element.type.folder:
            return element.thumbnail.s3_put_url

    class Meta(ElementTinySerializer.Meta):
        model = Element
        fields = ElementTinySerializer.Meta.fields + ('thumbnail_put_url',)
        read_only_fields = ElementTinySerializer.Meta.read_only_fields + ('thumbnail_put_url',)


class ElementListSerializer(ElementTinySerializer):
    """
    Serialize an full element with a specific array field 'best_classes'
    """
    best_classes = ClassificationSerializer(many=True, default=None, read_only=True)
    has_children = serializers.BooleanField(default=None, read_only=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if 'with_zone' in self.context and not self.context['with_zone']:
            self.fields.pop('zone')
        if 'with_corpus' in self.context and not self.context['with_corpus']:
            self.fields.pop('corpus')

    @extend_schema_field(serializers.CharField(allow_null=True))
    def get_thumbnail_url(self, element):
        # Do not return full element zone when with_zone query parameter is false, instead just return a thumbnail
        if 'with_zone' in self.context and not self.context['with_zone'] and not element.type.folder and element.zone:
            return element.zone.thumbnail_url
        return super().get_thumbnail_url(element)

    class Meta(ElementTinySerializer.Meta):
        model = Element
        fields = ElementTinySerializer.Meta.fields + ('best_classes', 'has_children', 'worker_version_id')


class ElementParentSerializer(serializers.Serializer):
    """
    Serialize a relation between an element and one of its parents
    """
    child = serializers.PrimaryKeyRelatedField(queryset=Element.objects.none())
    parent = serializers.PrimaryKeyRelatedField(queryset=Element.objects.none())

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):
            # Do not raise Error in order to create OpenAPI schema
            return
        corpora = Corpus.objects.writable(self.context['request'].user)
        self.fields['child'].queryset = Element.objects.filter(corpus__in=corpora)
        self.fields['parent'].queryset = Element.objects.filter(corpus__in=corpora)

    def validate(self, data):
        errors = defaultdict(list)
        data = super().validate(data)
        child = data.get('child')
        parent = data.get('parent')

        errors = defaultdict(list)
        if parent.corpus != child.corpus:
            errors['parent'].append("Parent is not from corpus '{}'".format(child.corpus.name))
        if parent.id == child.id:
            errors['parent'].append('A child cannot be its own parent')
        if errors:
            raise ValidationError(errors)
        return data

    def perform_create_checks(self, child, parent):
        # Assert parent is not an child's ancestor already
        if ElementPath.objects.filter(element_id=child.id, path__contains=[parent.id]).exists():
            raise ValidationError({'parent': [
                "'{}' is already a parent of '{}'".format(parent.id, child.id)
            ]})
        # Assert parent is not an alement's descendant
        if ElementPath.objects.filter(element_id=parent.id, path__contains=[child.id]).exists():
            raise ValidationError({'parent': [
                "'{}' is a child of element '{}'".format(parent.id, child.id)
            ]})

    def create(self, validated_data):
        child = validated_data['child']
        parent = validated_data['parent']
        self.perform_create_checks(child, parent)
        child.add_parent(parent)

    def delete(self, validated_data):
        child = validated_data['child']
        parent = validated_data['parent']
        parent.remove_child(child)


class ElementSerializer(ElementSlimSerializer):
    """
    Serialize an element with its metadata and classifications
    """

    metadata = serializers.SerializerMethodField()
    classifications = ClassificationSerializer(many=True, read_only=True)
    image = serializers.PrimaryKeyRelatedField(
        queryset=Image.objects.all().select_related('server'),
        required=False,
        write_only=True,
        help_text='Link this element to an image by UUID via a polygon. '
                  'When the image is updated, if there was an image before and the polygon is not updated, '
                  'the previous polygon is reused. Otherwise, a polygon filling the new image is used.'
    )
    polygon = LinearRingField(
        required=False,
        write_only=True,
        help_text='Set the polygon linking this element to the image. '
                  '`image` must be set when this field is set and there was no image or polygon defined before.',
    )

    class Meta:
        model = Element
        fields = ElementSlimSerializer.Meta.fields + (
            'metadata',
            'classifications',
            'image',
            'polygon',
            'worker_version'
        )
        read_only_fields = ElementSlimSerializer.Meta.read_only_fields + (
            'metadata',
            'classifications',
            'worker_version'
        )

    def create(self, *args, **kwargs):
        instance = super().create(*args, **kwargs)
        return instance

    def update(self, instance, validated_data):
        image = validated_data.pop('image', None)
        polygon = validated_data.pop('polygon', None)
        if polygon or image:
            if not image:
                if not instance.zone:
                    # A polygon was set but there is no image
                    raise ValidationError({
                        'image': ['Image is required when defining a polygon on an element without an existing zone'],
                    })
                image = instance.zone.image

            if image.width == 0 or image.height == 0:
                raise ValidationError({'image': ['This image does not have valid dimensions.']})

            if not polygon:
                if instance.zone:
                    polygon = instance.zone.polygon
                else:
                    polygon = LinearRing(
                        (0, 0),
                        (0, image.height),
                        (image.width, image.height),
                        (image.width, 0),
                        (0, 0)
                    )

            instance.zone, _ = image.zones.get_or_create(polygon=polygon)

        instance = super().update(instance, validated_data)
        return instance

    @extend_schema_field(MetaDataLightSerializer(many=True))
    def get_metadata(self, obj):
        if hasattr(self.instance, 'metadata_count'):
            return self.instance.metadata_count


class ElementNeighborsSerializer(serializers.Serializer):
    position = serializers.IntegerField(source='ordering')
    element = ElementLightSerializer()
    parents = serializers.ListField(
        child=ElementLightSerializer()
    )


class ElementCreateSerializer(ElementLightSerializer):
    """
    Serialize an Element with a possible parent and image
    """
    type = serializers.SlugField(source='type.slug')
    corpus = serializers.PrimaryKeyRelatedField(
        queryset=Corpus.objects.none(),
        style={'base_template': 'input.html'},
    )
    image = serializers.PrimaryKeyRelatedField(
        queryset=Image.objects.using('default').all().select_related('server'),
        required=False,
        help_text='Link this element to an image by UUID via a polygon.',
        style={'base_template': 'input.html'},
    )
    parent = serializers.PrimaryKeyRelatedField(
        queryset=Element.objects.none(),
        required=False,
        allow_null=True,
        style={'base_template': 'input.html'},
    )
    polygon = LinearRingField(
        required=False,
        help_text='Set the polygon linking this element to the image. '
                  '`image` must be set when this field is set. Defaults to a rectangle taking up the whole image.',
    )
    worker_version = serializers.PrimaryKeyRelatedField(
        queryset=WorkerVersion.objects.all(),
        required=False,
        allow_null=True,
        style={'base_template': 'input.html'},
    )

    class Meta(ElementLightSerializer.Meta):
        model = Element
        fields = ElementLightSerializer.Meta.fields + ('image', 'corpus', 'parent', 'polygon', 'worker_version')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):
            # Do not raise Error in order to create OpenAPI schema
            return
        corpora = Corpus.objects.writable(self.context['request'].user)
        self.fields['corpus'].queryset = corpora.prefetch_related('types')

        # Avoid a stale read issue when creating a child on a freshly created parent
        self.fields['parent'].queryset = Element.objects.filter(corpus__in=corpora).using('default')

    def validate(self, data):
        errors = defaultdict(list)
        data = super().validate(data)

        try:
            # Use types already loaded with the corpus queryset
            # instead of making another SQL query
            data['type'] = next(t for t in data['corpus'].types.all() if t.slug == data['type']['slug'])
        except StopIteration:
            # Fail immediately here because the type is required for other checks
            raise ValidationError({
                'type': ['Type {!r} not found in corpus {}'.format(data['type']['slug'], data['corpus'].name)]
            })

        parent = data.get('parent')
        if parent and parent.corpus_id != data['corpus'].id:
            errors['corpus'].append(
                'Parent and child must be in the same corpus'
            )
        image = data.get('image')
        if image and (image.width == 0 or image.height == 0):
            # Element creation with images with a width or height equal to zero
            # will lead to many errors everywhere as this would create impossible polygons
            errors['image'].append('This image does not have valid dimensions.')

        if errors:
            raise ValidationError(errors)
        return data

    def create(self, validated_data):
        image = validated_data.get('image')

        if image:
            # Create a new zone containing the image
            polygon = validated_data.get('polygon') or LinearRing(
                (0, 0),
                (0, image.height),
                (image.width, image.height),
                (image.width, 0),
                (0, 0)
            )

            # This is exactly like a Zone.get_or_create, but a bit more efficient:
            # - no savepoints as we do not need atomicity here (2 less queries)
            # - use .only('id') to load only the id
            #   This works because of the unique constraint on image_id + polygon
            try:
                zone = Zone \
                    .objects \
                    .using('default') \
                    .filter(
                        image_id=image.id,
                        polygon=polygon
                    ) \
                    .only('id') \
                    .get()
            except Zone.DoesNotExist:
                zone = image.zones.create(polygon=polygon)

            # Set the zone attributes to avoid retrieving the zone, image and server again
            # when the endpoint returns the created element without the slim_output mode.
            zone.image = image
            zone.polygon = polygon
        else:
            zone = None

        element = Element.objects.create(
            corpus=validated_data['corpus'],
            type=validated_data['type'],
            name=validated_data['name'],
            worker_version=validated_data.get('worker_version'),
            zone=zone,
        )

        # Make relation to parent, without dealing with children as there cannot be
        # any children on a newly created element
        parent = validated_data.get('parent')
        if parent:
            element.add_parent(parent, skip_children=True)

        return element

    def to_representation(self, instance):

        # Add a mode to only output the newly created element's id
        # This speeds up the endpoint execution time as it makes a few less SQL queries
        # and does not spend time serializing. This should be used by all workers
        if self.context['request'].query_params.get('slim_output') is not None:
            return {"id": instance.id}

        return ElementSerializer(instance, context=self.context).data


class ElementEntitiesSerializer(ElementLightSerializer):
    """
    Serialize an element with its entities
    Entities are split into two types, those extracted
    from metadata and those extracted from element transcriptions
    """
    metadata = MetaDataSerializer(many=True)
    transcriptions = TranscriptionEntityDetailsSerializer(many=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):
            # Keep original fields so OpenAPI can build its schemas
            return
        self.fields['transcriptions'] = serializers.SerializerMethodField()
        self.fields['metadata'] = serializers.SerializerMethodField()

    @cached_property
    def worker_version_id(self):
        if 'worker_version' not in self.context['request'].query_params:
            return

        try:
            worker_version_id = UUID(self.context['request'].query_params['worker_version'])
        except (TypeError, ValueError):
            raise ValidationError({'worker_version': ['Invalid UUID']})

        if not WorkerVersion.objects.filter(id=worker_version_id).exists():
            raise ValidationError({'worker_version': ['This worker version does not exist.']})

        return worker_version_id

    def get_transcriptions(self, element):
        filters = {
            'transcription__in': Transcription.objects.filter(element_id=element.id),
        }
        if self.worker_version_id:
            filters['entity__worker_version_id'] = self.worker_version_id

        transcription_entities = TranscriptionEntity.objects \
            .filter(**filters) \
            .prefetch_related('entity') \
            .order_by('offset')

        return TranscriptionEntityDetailsSerializer(transcription_entities, many=True).data

    @extend_schema_field(MetaDataSerializer(many=True))
    def get_metadata(self, obj):
        if self.worker_version_id:
            # entity__isnull is not needed here as this will already do an INNER JOIN and exclude nulls
            filters = {'entity__worker_version_id': self.worker_version_id}
        else:
            filters = {'entity__isnull': False}

        metadata = obj.metadatas \
            .filter(**filters) \
            .prefetch_related('entity')

        return MetaDataSerializer(metadata, many=True).data

    class Meta:
        model = Element
        fields = ('transcriptions', 'metadata')


class ElementTypeSerializer(ElementTypeLightSerializer):
    """
    Serialize an element type with its corpus
    """
    corpus = serializers.PrimaryKeyRelatedField(queryset=Corpus.objects.none())

    class Meta:
        model = ElementType
        fields = ElementTypeLightSerializer.Meta.fields + ('corpus', )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):
            # Keep original fields so OpenAPI can build its schemas
            return
        self.fields['corpus'].queryset = Corpus.objects.writable(self.context['request'].user)


class ElementBulkItemSerializer(serializers.Serializer):
    """
    A single element in a bulk element creation.
    All validation is done at once in ElementBulkSerializer. Do not use this serializer alone.
    """
    name = serializers.CharField()
    type = serializers.SlugField()
    polygon = LinearRingField()


class ElementBulkSerializer(serializers.Serializer):
    """
    Serialize child elements on a common parent
    """
    worker_version = serializers.PrimaryKeyRelatedField(
        queryset=WorkerVersion.objects.all(),
        required=False,
        allow_null=True,
        help_text='A WorkerVersion ID that the new elements will refer to',
        # Avoids thousands of SQL queries when opening the endpoint in a browser
        style={'base_template': 'input.html'},
    )
    elements = ElementBulkItemSerializer(
        many=True,
        allow_empty=False,
        help_text='A list of elements with their name, type and polygon'
    )

    def validate(self, data):
        type_slugs = set(element['type'] for element in data['elements'])
        type_ids = dict(
            ElementType
            .objects
            .filter(
                corpus_id=self.context['element'].corpus_id,
                slug__in=type_slugs)
            .values_list('slug', 'id')
        )
        missing_types = list(type_slugs - set(type_ids.keys()))
        if missing_types:
            raise ValidationError([
                f"Element types with slugs {', '.join(missing_types)} do not exist in the parent element's corpus"
            ])

        for element in data['elements']:
            # Replace slugs with type IDs
            element['type'] = type_ids[element['type']]

        return data
