import uuid
from unittest.mock import call, patch

from django.db.models import Count
from django.test import override_settings
from django.urls import reverse
from rest_framework import status

from arkindex.dataimport.models import WorkerVersion
from arkindex.documents.models import Corpus, Element
from arkindex.project.tests import FixtureAPITestCase


class TestBulkElementTranscriptions(FixtureAPITestCase):
    """
    Tests for text element creation view
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.page = cls.corpus.elements.get(name='Volume 1, page 1r')
        cls.vol = cls.corpus.elements.get(name='Volume 1')
        cls.line = cls.corpus.elements.filter(type__slug='text_line').first()
        cls.private_corpus = Corpus.objects.create(name='Private')
        cls.private_page = cls.private_corpus.elements.create(type=cls.page.type)
        cls.worker_version = WorkerVersion.objects.get(worker__slug='reco')

    @patch('arkindex.project.triggers.tasks.reindex_start.delay')
    def test_bulk_transcriptions(self, delay_mock):
        """
        Bulk creates a list of element with an attached transcription generated by a worker_version
        """
        # Create a manual transcription on the element
        self.line.transcriptions.create(
            text='A manual transcription',
            worker_version=self.worker_version,
        )
        self.assertEqual(self.line.transcriptions.count(), 1)

        existing_element_ids = list(Element.objects.get_descending(self.page.id).values_list('id', flat=True))

        transcriptions = [
            ([[13, 37], [133, 37], [133, 137], [13, 137], [13, 37]], 'Hello world !', 0.1337),
            ([[24, 42], [64, 42], [64, 142], [24, 142], [24, 42]], 'I <3 JavaScript', 0.42),
        ]
        data = {
            'element_type': 'text_line',
            'worker_version': str(self.worker_version.id),
            'transcriptions': [{
                'polygon': poly,
                'text': text,
                'confidence': confidence
            } for poly, text, confidence in transcriptions]
        }
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:element-transcriptions-bulk', kwargs={'pk': self.page.id}),
            format='json',
            data=data
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        created_elts = Element.objects.get_descending(self.page.id).exclude(id__in=existing_element_ids)
        self.assertEqual(created_elts.count(), 2)
        self.assertTrue(all(map(lambda elt: elt.zone.image == self.page.zone.image, created_elts)))
        self.assertListEqual(
            [
                (elt.paths.first().ordering, elt.name, elt.zone.polygon.coords)
                for elt in created_elts
            ],
            [
                (1, '2', ((13, 37), (13, 137), (133, 137), (133, 37), (13, 37))),
                (2, '3', ((24, 42), (24, 142), (64, 142), (64, 42), (24, 42)))
            ]
        )
        self.assertCountEqual(
            created_elts.values_list('transcriptions__text', 'transcriptions__worker_version'),
            [
                ('Hello world !', self.worker_version.id),
                ('I <3 JavaScript', self.worker_version.id)
            ]
        )
        self.assertEqual(delay_mock.call_count, 1)
        self.assertEqual(delay_mock.call_args, call(
            element_id=str(self.page.id),
            corpus_id=None,
            entity_id=None,
            transcriptions=True,
            elements=True,
            entities=False,
            drop=False,
            description=f'Indexation of element {self.page.id}',
            user_id=None,
        ))

    @override_settings(ARKINDEX_FEATURES={'search': False})
    def test_bulk_transcriptions_load(self):
        """
        Sending 100 transcriptions should not increase the number of DB queries
        """
        # Create 100 transcriptions
        transcriptions = [(
            [[i, i], [i, i + 20], [i + 20, i + 20], [i + 20, i], [i, i]],
            str(i / 2),
            i / 200
        ) for i in range(1, 200, 2)]
        data = {
            'element_type': 'text_line',
            'worker_version': str(self.worker_version.id),
            'transcriptions': [{
                'polygon': poly,
                'text': text,
                'confidence': confidence
            } for poly, text, confidence in transcriptions]
        }

        created_elts = Element.objects.get_descending(self.page.id) \
            .filter(paths__path__last=self.page.id, type__slug='text_line')
        self.assertEqual(created_elts.count(), 1)

        self.client.force_login(self.internal_user)
        with self.assertNumQueries(16):
            response = self.client.post(
                reverse('api:element-transcriptions-bulk', kwargs={'pk': self.page.id}),
                format='json',
                data=data
            )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(created_elts.count(), 101)
        self.assertEqual(created_elts.count(), 101)
        # Each annotated element has a transcription
        self.assertEqual(created_elts.annotate(ts_count=Count('transcriptions')).filter(ts_count=1).count(), 100)

    @patch('arkindex.project.triggers.tasks.reindex_start.delay')
    def test_bulk_transcriptions_similar_zone(self, delay_mock):
        """
        Pushing a transcription matching an element type and zone reuses this element
        Does not erase present transcriptions, even from the same worker
        """
        # Create a manual transcription on the element
        self.line.transcriptions.create(
            text='A manual transcription',
            worker_version=self.worker_version,
        )
        self.assertEqual(self.line.transcriptions.count(), 1)

        existing_element_ids = list(Element.objects.get_descending(self.page.id).values_list('id', flat=True))
        transcriptions = [
            ([[13, 37], [133, 37], [133, 137], [13, 137], [13, 37]], 'Hello world !', 0.1337),
            # Use line zone to create the second transcription
            (list(self.line.zone.polygon), 'I <3 JavaScript', 0.42),
        ]
        data = {
            'element_type': 'text_line',
            'worker_version': str(self.worker_version.id),
            'transcriptions': [{
                'polygon': poly,
                'text': text,
                'confidence': confidence
            } for poly, text, confidence in transcriptions]
        }

        self.client.force_login(self.internal_user)
        with self.assertNumQueries(16):
            response = self.client.post(
                reverse('api:element-transcriptions-bulk', kwargs={'pk': self.page.id}),
                format='json',
                data=data
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        created_elts = Element.objects.get_descending(self.page.id).exclude(id__in=existing_element_ids)
        # The existing text line has been reused
        self.assertEqual(created_elts.count(), 1)
        self.assertCountEqual(self.line.transcriptions.values_list('text', flat=True), [
            'A manual transcription',
            'I <3 JavaScript'
        ])

    def test_bulk_transcriptions_requires_verified(self):
        """
        The user has to be verified tu push transcriptions in bulk
        """
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:element-transcriptions-bulk', kwargs={'pk': self.page.id}),
            format='json',
            data={}
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have permission to perform this action.'})

    def test_bulk_transcriptions_private(self):
        """
        Accessing a non writeable element triggers a 404_NOT_FOUND
        """
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:element-transcriptions-bulk', kwargs={'pk': self.private_page.id}),
            format='json',
            data={}
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_bulk_transcriptions_non_manual(self):
        """
        Worker version is a required field
        """
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:element-transcriptions-bulk', kwargs={'pk': self.page.id}),
            format='json',
            data={
                'element_type': 'text_line',
                'transcriptions': [{
                    'polygon': [[13, 37], [133, 37], [133, 137], [13, 137], [13, 37]],
                    'text': 'Can I write womething here ?',
                    'confidence': .666
                }]
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'worker_version': ['This field is required.']
        })

    def test_bulk_transcriptions_wrong_fields_validation(self):
        """
        Briefly assert fields values are validated
        """
        wrong_version_id = str(uuid.uuid4())
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:element-transcriptions-bulk', kwargs={'pk': self.page.id}),
            format='json',
            data={
                'element_type': 'paragraph',
                'worker_version': wrong_version_id,
                'transcriptions': [
                    {
                        'polygon': [[13, 37]],
                        'text': 'There is a snake in my computer'
                    },
                    {
                        'polygon': [[13, 37], [133, 37], [133, 137], [13, 137], [13, 37]],
                        'text': 'No.',
                    },
                    {
                        'polygon': [[13, 37], [133, 37], [133, 137], [13, 137], [13, 37]],
                        'text': 'How about both',
                        'score': .42,
                        'confidence': .1337
                    }
                ]
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'element_type': ['Object with slug=paragraph does not exist.'],
            'worker_version': [f'Invalid pk "{wrong_version_id}" - object does not exist.'],
            'transcriptions': [
                {
                    'polygon': ['Ensure this field has at least 3 elements.']
                },
                {
                    'non_field_errors': ['Either the score or confidence field should be defined.'],
                },
                {
                    'non_field_errors': ['Either the score or confidence field should be defined.'],
                },
            ]
        })

    @override_settings(ARKINDEX_FEATURES={'search': False})
    def test_bulk_transcriptions_return_ordered_elements(self):
        """
        return_elements field allows to serialize an ordered list of sub-elements holding the transcriptions
        """
        transcriptions = [
            ([[13, 37], [133, 37], [133, 137], [13, 137], [13, 37]], 'stuck', 0.13),
            ([[24, 42], [64, 42], [64, 142], [24, 142], [24, 42]], 'stock', 0.37),
            # The existing line has the smallest ordering
            (self.line.zone.polygon.coords, 'stack', 0.42),
        ]
        data = {
            'element_type': 'text_line',
            'worker_version': str(self.worker_version.id),
            'transcriptions': [{
                'polygon': poly,
                'text': text,
                'confidence': confidence
            } for poly, text, confidence in transcriptions],
            'return_elements': True
        }

        self.client.force_login(self.internal_user)
        with self.assertNumQueries(16):
            response = self.client.post(
                reverse('api:element-transcriptions-bulk', kwargs={'pk': self.page.id}),
                format='json',
                data=data
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        created_elts = Element.objects.get_descending(self.page.id).exclude(id=self.line.id)
        self.assertEqual(created_elts.count(), 5)

        first_elt = created_elts.get(transcriptions__text='stuck')
        second_elt = created_elts.get(transcriptions__text='stock')
        self.assertListEqual(response.json(), [{
            'id': str(first_elt.id),
            'created': True
        }, {
            'id': str(second_elt.id),
            'created': True
        }, {
            'id': str(self.line.id),
            'created': False
        }])

    def test_bulk_transcriptions_requires_zone(self):
        self.client.force_login(self.internal_user)
        response = self.client.post(
            reverse('api:element-transcriptions-bulk', kwargs={'pk': self.vol.id}),
            format='json',
            data={}
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'element': ['Cannot create transcriptions on an element without a zone.']})

    @patch('arkindex.project.triggers.tasks.reindex_start.delay')
    def test_top_level_element(self, delay_mock):
        """
        Create transcriptions on a top level element
        """
        # Create a top level page
        top_level = self.corpus.elements.create(
            type=self.page.type,
            name='Top level page',
            zone=self.page.zone,
        )

        transcriptions = [
            ([[13, 37], [133, 37], [133, 137], [13, 137], [13, 37]], 'Hello world !', 0.1337),
            ([[24, 42], [64, 42], [64, 142], [24, 142], [24, 42]], 'I <3 JavaScript', 0.42),
        ]
        data = {
            'element_type': 'text_line',
            'worker_version': str(self.worker_version.id),
            'transcriptions': [{
                'polygon': poly,
                'text': text,
                'confidence': confidence
            } for poly, text, confidence in transcriptions]
        }
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:element-transcriptions-bulk', kwargs={'pk': top_level.id}),
            format='json',
            data=data
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        created_elts = Element.objects.get_descending(top_level.id)
        self.assertEqual(created_elts.count(), 2)
        self.assertTrue(all(map(lambda elt: elt.zone.image == self.page.zone.image, created_elts)))
        self.assertListEqual(
            [
                (elt.paths.first().ordering, elt.name, elt.zone.polygon.coords)
                for elt in created_elts
            ],
            [
                (0, '1', ((13, 37), (13, 137), (133, 137), (133, 37), (13, 37))),
                (1, '2', ((24, 42), (24, 142), (64, 142), (64, 42), (24, 42)))
            ]
        )
        self.assertCountEqual(
            created_elts.values_list('transcriptions__text', 'transcriptions__worker_version'),
            [
                ('Hello world !', self.worker_version.id),
                ('I <3 JavaScript', self.worker_version.id)
            ]
        )
        self.assertEqual(delay_mock.call_count, 1)
        self.assertEqual(delay_mock.call_args, call(
            element_id=str(top_level.id),
            corpus_id=None,
            entity_id=None,
            transcriptions=True,
            elements=True,
            entities=False,
            drop=False,
            user_id=None,
            description=f'Indexation of element {top_level.id}'
        ))
