import uuid

from django.urls import reverse
from rest_framework import status

from arkindex.dataimport.models import WorkerVersion
from arkindex.documents.models import Element
from arkindex.project.tests import FixtureAPITestCase


class TestChildrenElements(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.vol = cls.corpus.elements.get(name='Volume 1')
        cls.worker_version = WorkerVersion.objects.get(worker__slug='dla')

    def test_element_children(self):
        lonely_element = Element.objects.get(name='Volume 2, page 2r')
        response = self.client.get(reverse('api:elements-children', kwargs={'pk': str(lonely_element.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json()['results'], [])

        response = self.client.get(reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.json()['results']), 8)
        self.assertSetEqual(
            {r['type'] for r in response.json()['results']},
            {'act', 'page'},
        )

        response = self.client.get(
            reverse('api:elements-children', kwargs={'pk': str(self.vol.id)})
            + '?recursive&type=surface'
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertCountEqual(
            [r['name'] for r in response.json()['results']],
            [
                'Surface A',
                'Surface B',
                'Surface C',
                'Surface D',
                'Surface E',
                'Surface F'
            ],
        )

        response = self.client.get(reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}) + '?recursive')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.json()['results']), 20)
        self.assertSetEqual(
            {r['type'] for r in response.json()['results']},
            {'act', 'page', 'surface', 'text_line', 'word'},
        )

        response = self.client.get(reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}) + '?type=page')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [r['name'] for r in response.json()['results']],
            [
                'Volume 1, page 1r',
                'Volume 1, page 1v',
                'Volume 1, page 2r'
            ]
        )

        response = self.client.get(reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}) + '?name=2')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [r['name'] for r in response.json()['results']],
            [
                'Act 2',
                'Volume 1, page 2r',
            ]
        )

        nested_volume = self.corpus.elements.create(
            type=self.vol.type,
            name='Nested volume',
        )
        nested_volume.add_parent(self.vol)

        response = self.client.get(reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}) + '?folder')
        self.assertListEqual(response.json()['results'], [
            {
                "id": str(nested_volume.id),
                "name": nested_volume.name,
                "type": nested_volume.type.slug,
                "worker_version_id": None,
                "zone": None,
                "corpus": {
                    "id": str(self.corpus.id),
                    "name": self.corpus.name,
                    "public": self.corpus.public,
                },
                "best_classes": None,
                "has_children": None,
                "thumbnail_url": nested_volume.thumbnail.s3_url,
            }
        ])

        nested_volume.worker_version = self.worker_version
        nested_volume.save()
        response = self.client.get(reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}) + '?folder')
        self.assertListEqual(response.json()['results'], [
            {
                "id": str(nested_volume.id),
                "name": nested_volume.name,
                "type": nested_volume.type.slug,
                "worker_version_id": str(self.worker_version.id),
                "zone": None,
                "corpus": {
                    "id": str(self.corpus.id),
                    "name": self.corpus.name,
                    "public": self.corpus.public,
                },
                "best_classes": None,
                "has_children": None,
                "thumbnail_url": nested_volume.thumbnail.s3_url,
            }
        ])

    def test_element_children_with_zone_true(self):
        lonely_element = Element.objects.get(name='Volume 1, page 2r')
        response = self.client.get(
            reverse('api:elements-children', kwargs={'pk': str(self.vol.id)})
            + '?with_zone=true&type=page&name=2'
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json()['results'], [
            {
                "id": str(lonely_element.id),
                "name": lonely_element.name,
                "type": lonely_element.type.slug,
                "worker_version_id": None,
                "zone": {
                    "id": str(lonely_element.zone.id),
                    "image" : {
                        "id": str(lonely_element.zone.image.id),
                        "height": lonely_element.zone.image.height,
                        "width": lonely_element.zone.image.width,
                        "path": lonely_element.zone.image.path,
                        "s3_url": None,
                        "server": {
                            "display_name": lonely_element.zone.image.server.display_name,
                            "max_height": None,
                            "max_width": None,
                            "url": lonely_element.zone.image.server.url,
                        },
                        "status": lonely_element.zone.image.status.value,
                        "url": lonely_element.zone.image.url
                    },
                    "polygon": [[0, 0], [0, 1000], [1000, 1000], [1000, 0], [0, 0]],
                    "url": lonely_element.zone.url,
                },
                "corpus": {
                    "id": str(self.corpus.id),
                    "name": self.corpus.name,
                    "public": self.corpus.public,
                },
                "best_classes": None,
                "has_children": None,
                "thumbnail_url": None,
            }
        ])

    def test_element_children_with_zone_false(self):
        lonely_element = Element.objects.get(name='Volume 1, page 2r')
        response = self.client.get(
            reverse('api:elements-children', kwargs={'pk': str(self.vol.id)})
            + '?with_zone=false&type=page&name=2'
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json()['results'], [
            {
                "id": str(lonely_element.id),
                "name": lonely_element.name,
                "type": lonely_element.type.slug,
                "worker_version_id": None,
                "corpus": {
                    "id": str(self.corpus.id),
                    "name": self.corpus.name,
                    "public": self.corpus.public,
                },
                "best_classes": None,
                "has_children": None,
                "thumbnail_url": lonely_element.zone.thumbnail_url,
            }
        ])

    def test_element_children_with_corpus_false(self):
        lonely_element = Element.objects.get(name='Volume 1, page 2r')
        response = self.client.get(
            reverse('api:elements-children', kwargs={'pk': str(self.vol.id)})
            + '?with_corpus=false&type=page&name=2'
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json()['results'], [
            {
                "id": str(lonely_element.id),
                "name": lonely_element.name,
                "type": lonely_element.type.slug,
                "worker_version_id": None,
                "zone": {
                    "id": str(lonely_element.zone.id),
                    "image" : {
                        "id": str(lonely_element.zone.image.id),
                        "height": lonely_element.zone.image.height,
                        "width": lonely_element.zone.image.width,
                        "path": lonely_element.zone.image.path,
                        "s3_url": None,
                        "server": {
                            "display_name": lonely_element.zone.image.server.display_name,
                            "max_height": None,
                            "max_width": None,
                            "url": lonely_element.zone.image.server.url,
                        },
                        "status": lonely_element.zone.image.status.value,
                        "url": lonely_element.zone.image.url
                    },
                    "polygon": [[0, 0], [0, 1000], [1000, 1000], [1000, 0], [0, 0]],
                    "url": lonely_element.zone.url,
                },
                "best_classes": None,
                "has_children": None,
                "thumbnail_url": None,
            }
        ])

    def test_element_children_worker_version(self):
        self.corpus.elements.filter(name__contains='page 1r').update(worker_version=self.worker_version)

        with self.assertNumQueries(7):
            response = self.client.get(
                reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}),
                data={'worker_version': str(self.worker_version.id)}
            )

        self.assertListEqual(
            [r['name'] for r in response.json()['results']],
            [
                'Volume 1, page 1r',
            ]
        )

    def test_element_children_worker_version_validation(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}),
                data={'worker_version': 'blah'},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'worker_version': ['Invalid UUID']})

        with self.assertNumQueries(2):
            response = self.client.get(
                reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}),
                data={'worker_version': uuid.uuid4()},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'worker_version': ['This worker version does not exist.']})

    def test_element_children_ordering(self):
        expected = [
            'Act 4',
            'Act 5',
            'Act 1',
            'Act 2',
            'Act 3'
        ]
        for ordering, name in enumerate(expected):
            Element.objects.get(type__slug='act', name=name).paths.update(ordering=ordering)

        response = self.client.get(
            reverse('api:elements-children', kwargs={'pk': str(self.vol.id)})
            + '?recursive&type=act'
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [r['name'] for r in response.json()['results']],
            expected,
        )

    def test_children_with_has_children(self):
        # Create a childless element
        element = self.corpus.elements.create(
            name='Spearow',
            type=self.corpus.types.get(slug='surface')
        )
        element.add_parent(self.vol)

        with self.assertNumQueries(8):
            response = self.client.get(
                reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}),
                data={'with_has_children': True},
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(
            {element['name']: element['has_children'] for element in response.json()['results']},
            {
                'Act 1': True,
                'Act 2': True,
                'Act 3': True,
                'Act 4': True,
                'Act 5': True,
                'Volume 1, page 1r': True,
                'Volume 1, page 1v': True,
                'Volume 1, page 2r': True,
                'Spearow': False,
            }
        )

    def test_children_modified_since_bad_format(self):
        with self.assertNumQueries(2):
            response = self.client.get(
                reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}),
                HTTP_IF_MODIFIED_SINCE='AAAAAAAAAAAAAAAAA',
            )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'If-Modified-Since': ['Bad date format']})

    def test_children_modified_since_not_modified(self):
        with self.assertNumQueries(3):
            response = self.client.get(
                reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}),
                HTTP_IF_MODIFIED_SINCE='Thu, 02 Apr 2099 13:37:42 GMT',
            )
        self.assertEqual(response.status_code, status.HTTP_304_NOT_MODIFIED)

    def test_children_modified_since(self):
        self.corpus.elements.filter(name='Volume 1, page 1r').update(updated='2099-04-02T13:37:43Z')
        with self.assertNumQueries(7):
            response = self.client.get(
                reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}),
                HTTP_IF_MODIFIED_SINCE='Thu, 02 Apr 2099 13:37:42 GMT',
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 8)
        self.assertSetEqual(
            {r['type'] for r in data['results']},
            {'act', 'page'},
        )
