from unittest.mock import call, patch

from django.urls import reverse
from rest_framework import status

from arkindex.dataimport.models import WorkerVersion
from arkindex.documents.models import Corpus
from arkindex.project.tests import FixtureAPITestCase


class TestDestroyWorkerResults(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.version_1 = WorkerVersion.objects.get(worker__slug='reco')
        cls.page1 = cls.corpus.elements.get(name='Volume 1, page 2r')
        cls.private_corpus = Corpus.objects.create(name='private', public=False)

    def test_worker_results_destroy_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.delete(reverse('api:worker-delete-results', kwargs={'corpus': str(self.corpus.id), 'version': str(self.version_1.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_worker_results_destroy_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        with self.assertNumQueries(0):
            response = self.client.delete(reverse('api:worker-delete-results', kwargs={'corpus': str(self.corpus.id), 'version': str(self.version_1.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_worker_results_destroy_wrong_acl(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.delete(reverse('api:worker-delete-results', kwargs={'corpus': str(self.private_corpus.id), 'version': str(self.version_1.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(
            response.json(),
            {'detail': 'You do not have admin access to this corpus.'}
        )

    def test_worker_results_destroy_wrong_corpus_id(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.delete(reverse('api:worker-delete-results', kwargs={'corpus': '12341234-1234-1234-1234-123412341234', 'version': str(self.version_1.id)}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertDictEqual(
            response.json(),
            {'detail': 'Not found.'}
        )

    def test_worker_results_destroy_wrong_version_id(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.delete(reverse('api:worker-delete-results', kwargs={'corpus': str(self.corpus.id), 'version': '12341234-1234-1234-1234-123412341234'}))
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'worker_version_id': ['This version does not exist.']}
        )

    @patch('arkindex.project.triggers.tasks.worker_results_delete.delay')
    def test_worker_results_destroy_on_corpus(self, delay_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.delete(reverse('api:worker-delete-results', kwargs={'corpus': str(self.corpus.id), 'version': str(self.version_1.id)}))
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(delay_mock.call_count, 1)
        self.assertEqual(delay_mock.call_args, call(
            corpus_id=self.corpus.id,
            version_id=self.version_1.id,
            parent_id=None,
            user_id=self.user.id,
            description=f"Deletion of results produced by worker version: {self.version_1}"
        ))

    def test_worker_results_destroy_invalid_element_id(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.delete(reverse('api:worker-delete-results', kwargs={'corpus': str(self.corpus.id), 'version': str(self.version_1.id)}) + '?element_id=pouet')
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'element_id': ['Invalid UUID.']}
        )

    def test_worker_results_destroy_wrong_element_id(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.delete(reverse('api:worker-delete-results', kwargs={'corpus': str(self.corpus.id), 'version': str(self.version_1.id)}) + '?element_id=12341234-1234-1234-1234-123412341234')
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'element_id': ['This element does not exist.']}
        )

    @patch('arkindex.project.triggers.tasks.worker_results_delete.delay')
    def test_worker_results_destroy_on_parent_element(self, delay_mock):
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.delete(reverse('api:worker-delete-results', kwargs={'corpus': str(self.corpus.id), 'version': str(self.version_1.id)}) + f'?element_id={self.page1.id}')
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(delay_mock.call_count, 1)
        self.assertEqual(delay_mock.call_args, call(
            corpus_id=self.corpus.id,
            version_id=self.version_1.id,
            parent_id=self.page1.id,
            user_id=self.user.id,
            description=f"Deletion of results produced by worker version: {self.version_1}"
        ))
