from unittest.mock import MagicMock, call, patch

from elasticsearch import Elasticsearch
from elasticsearch.exceptions import NotFoundError

from arkindex.dataimport.models import WorkerVersion
from arkindex.documents.indexer import Indexer
from arkindex.documents.models import EntityType
from arkindex.project.tests import FixtureTestCase


class TestIndexer(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        worker_version = WorkerVersion.objects.first()
        for i in range(10):
            cls.corpus.entities.create(
                name=f'ES Dummy {i}',
                type=EntityType.Misc,
                worker_version=worker_version,
            )

    @patch('arkindex.documents.indexer.Elasticsearch')
    def test_drop_index(self, es_mock):
        document_mock = MagicMock()
        document_mock.__name__ = 'ESDummy'
        indexer = Indexer(['a', 'b'])
        self.assertEqual(es_mock.call_count, 1)
        self.assertEqual(es_mock.call_args, call(['a', 'b']))

        indexer.drop_index(document_mock)
        self.assertEqual(document_mock._index.delete.call_count, 1)
        self.assertEqual(document_mock._index.delete.call_args, call())

    @patch('arkindex.documents.indexer.Elasticsearch')
    def test_drop_index_handle_exceptions(self, es_mock):
        # Catches NotFoundError
        document_mock = MagicMock()
        document_mock.__name__ = 'ESDummy'
        document_mock._index.delete.side_effect = NotFoundError
        Indexer().drop_index(document_mock)
        self.assertEqual(document_mock._index.delete.call_count, 1)
        self.assertEqual(document_mock._index.delete.call_args, call())

        # Raises anything else
        document_mock.reset_mock()
        document_mock._index.delete.side_effect = ValueError
        with self.assertRaises(ValueError):
            Indexer().drop_index(document_mock)
        self.assertEqual(document_mock._index.delete.call_count, 1)
        self.assertEqual(document_mock._index.delete.call_args, call())

    @patch('arkindex.documents.indexer.es_bulk')
    def test_run_index(self, es_bulk_mock):
        es_bulk_mock.return_value = (5, [])

        queryset = self.corpus.entities.filter(name__startswith='ES Dummy')
        Indexer().run_index(queryset, bulk_size=5)

        self.assertEqual(es_bulk_mock.call_count, 2)
        for i, ((es_instance, actions), kwargs) in enumerate(es_bulk_mock.call_args_list):
            self.assertIsInstance(es_instance, Elasticsearch)
            self.assertListEqual(actions, [
                {
                    '_index': 'entities',
                    '_type': 'entity',
                    '_id': entity.id.hex,
                    '_source': {
                        'corpus': self.corpus.id,
                        'type': entity.type.value,
                        'name': entity.name,
                    },
                }
                for entity in queryset[i * 5:(i + 1) * 5]
            ])
            self.assertDictEqual(kwargs, {
                'chunk_size': 100,
                'initial_backoff': 2,
                'max_retries': 2,
                'request_timeout': 15,
                'stats_only': True,
                '_source': False,
            })

    @patch('arkindex.documents.indexer.es_bulk')
    def test_index_retry(self, es_bulk_mock):
        # Keep failing
        es_bulk_mock.side_effect = Exception

        queryset = self.corpus.entities.filter(name__startswith='ES Dummy')
        count = Indexer().index(queryset, chunk_size=10, timeout=42)

        self.assertEqual(count, 0)
        self.assertEqual(es_bulk_mock.call_count, 3)
        args, kwargs = zip(*es_bulk_mock.call_args_list)

        for es_instance, actions in args:
            self.assertIsInstance(es_instance, Elasticsearch)
            self.assertListEqual(actions, [
                {
                    '_index': 'entities',
                    '_type': 'entity',
                    '_id': entity.id.hex,
                    '_source': {
                        'corpus': self.corpus.id,
                        'type': entity.type.value,
                        'name': entity.name,
                    },
                }
                for entity in queryset
            ])

        self.assertTupleEqual(kwargs, (
            {
                'chunk_size': 10,
                'initial_backoff': 2,
                'max_retries': 2,
                'request_timeout': 42,
                'stats_only': True,
                '_source': False,
            },
            {
                'chunk_size': 5,
                'initial_backoff': 2,
                'max_retries': 2,
                'request_timeout': 84,
                'stats_only': True,
                '_source': False,
            },
            {
                'chunk_size': 2,
                'initial_backoff': 2,
                'max_retries': 2,
                'request_timeout': 168,
                'stats_only': True,
                '_source': False,
            },
        ))
