import uuid
from enum import Enum
from pathlib import Path

from teklia_toolbox.config import ConfigParser, ConfigurationError, dir_path


class CacheType(Enum):
    Redis = 'redis'
    Memcached = 'memcached'
    Filesystem = 'filesystem'
    Memory = 'memory'
    Dummy = 'dummy'


class CookieSameSiteOption(Enum):
    """
    Options for the SameSite flag on a cookie. Django accepts Lax, Strict, None and False.
    None is a 'None' string, not Python's None, and disables the SameSite protection.
    This can cause warnings when the Secure flag is active.
    False removes the flag entirely from the cookie, leaving the decision up to the browser, which can cause warnings.

    https://docs.djangoproject.com/en/3.1/ref/settings/#std:setting-SESSION_COOKIE_SAMESITE
    """
    Lax = 'lax'
    Strict = 'strict'
    # Cannot redefine Python's None!
    None_ = 'none'
    Disabled = False


def get_settings_parser(base_dir):
    parser = ConfigParser()
    parser.add_option('arkindex_env', type=str, default='dev')
    parser.add_option('internal_group_id', type=int, default=2)
    parser.add_option('local_imageserver_id', type=int, default=1)
    parser.add_option('allowed_hosts', type=str, many=True, default=[])
    parser.add_option('imports_worker_version', type=uuid.UUID, default=None)

    # SECURITY WARNING: keep the secret key used in production secret!
    parser.add_option('secret_key', type=str, default='jf0w^y&ml(caax8f&a1mub)(js9(l5mhbbhosz3gi+m01ex+lo')
    parser.add_option('jwt_signing_key', type=str, default=None)

    database_parser = parser.add_subparser('database', default={})
    database_parser.add_option('name', type=str, default='arkindex_dev')
    database_parser.add_option('host', type=str, default='localhost')
    database_parser.add_option('port', type=int, default=9100)
    database_parser.add_option('user', type=str, default='devuser')
    database_parser.add_option('password', type=str, default='devdata')

    replica_parser = database_parser.add_subparser('replica', default=None)
    replica_parser.add_option('name', type=str, default=None)
    replica_parser.add_option('host', type=str, default=None)
    replica_parser.add_option('port', type=int, default=None)
    replica_parser.add_option('user', type=str, default=None)
    replica_parser.add_option('password', type=str, default=None)

    email_parser = parser.add_subparser('email', default=None)
    email_parser.add_option('host', type=str)
    email_parser.add_option('port', type=int)
    email_parser.add_option('user', type=str)
    email_parser.add_option('password', type=str)
    email_parser.add_option('error_report_recipients', type=str, many=True, default=[])

    static_parser = parser.add_subparser('static', default={})
    static_parser.add_option('root_path', type=dir_path, default=None)
    static_parser.add_option('cdn_assets_url', type=str, default=None)
    static_parser.add_option('mirador_url', type=str, default=None)
    static_parser.add_option('universal_viewer_url', type=str, default=None)
    static_parser.add_option('frontend_version', type=str, default=None)

    elasticsearch_parser = parser.add_subparser('elasticsearch', default={})
    elasticsearch_parser.add_option('hosts', type=str, many=True, default=['localhost'])

    influxdb_parser = parser.add_subparser('influxdb', default={})
    influxdb_parser.add_option('api_url', type=str, default='http://localhost:8086/')

    transkribus_parser = parser.add_subparser('transkribus', default={})
    transkribus_parser.add_option('email', type=str, default=None)
    transkribus_parser.add_option('password', type=str, default=None)
    transkribus_parser.add_option('worker_version', type=uuid.UUID, default=None)

    gitlab_parser = parser.add_subparser('gitlab', default={})
    gitlab_parser.add_option('app_id', type=str, default=None)
    gitlab_parser.add_option('app_secret', type=str, default=None)

    redis_parser = parser.add_subparser('redis', default={})
    redis_parser.add_option('host', type=str, default='localhost')
    redis_parser.add_option('port', type=int, default=6379)
    redis_parser.add_option('db', type=int, default=0)
    redis_parser.add_option('password', type=str, default=None)
    redis_parser.add_option('timeout', type=int, default=1800)

    csrf_parser = parser.add_subparser('csrf', default={})
    csrf_parser.add_option('cookie_name', type=str, default='arkindex.csrf')
    csrf_parser.add_option('cookie_domain', type=str, default=None)
    csrf_parser.add_option('cookie_secure', type=bool, default=False)
    csrf_parser.add_option('cookie_samesite', type=CookieSameSiteOption, default=CookieSameSiteOption.Lax)
    csrf_parser.add_option('trusted_origins', type=str, many=True, default=[])

    session_parser = parser.add_subparser('session', default={})
    session_parser.add_option('cookie_name', type=str, default='arkindex.auth')
    session_parser.add_option('cookie_domain', type=str, default=None)
    session_parser.add_option('cookie_secure', type=bool, default=False)
    session_parser.add_option('cookie_samesite', type=CookieSameSiteOption, default=CookieSameSiteOption.Lax)

    cors_parser = parser.add_subparser('cors', default={})
    cors_parser.add_option('origin_whitelist', type=str, many=True, default=[
        'http://localhost:8080',
        'http://127.0.0.1:8080',
    ])
    cors_parser.add_option('suffixes', type=str, many=True, default=[])

    ponos_parser = parser.add_subparser('ponos', default={})
    # Do not use file_path here to allow the backend to start without a Ponos key
    ponos_parser.add_option('private_key', type=Path, default=(base_dir / 'ponos.key').resolve())
    ponos_parser.add_option('default_env', type=dict, default={})
    ponos_parser.add_option('default_farm', type=uuid.UUID, default=None)

    docker_parser = parser.add_subparser('docker', default={})
    docker_parser.add_option('tasks_image', type=str, default='registry.gitlab.com/arkindex/tasks')

    sentry_parser = parser.add_subparser('sentry', default={})
    sentry_parser.add_option('dsn', type=str, default=None)
    sentry_parser.add_option('frontend_dsn', type=str, default=None)

    features_parser = parser.add_subparser('features', allow_extra_keys=False, default={})
    features_parser.add_option('signup', type=bool, default=True)
    features_parser.add_option('selection', type=bool, default=True)
    features_parser.add_option('search', type=bool, default=True)
    features_parser.add_option('transkribus', type=bool, default=True)
    features_parser.add_option('workers', type=bool, default=False)
    features_parser.add_option('workers_activity', type=bool, default=False)

    cache_parser = ConfigParser()
    cache_parser.add_option('type', type=CacheType, default=None)
    cache_parser.add_option('url', type=str, default=None)
    cache_parser.add_option('path', type=dir_path, default=None)

    def cache_validator(value):
        data = cache_parser.parse_data(value)
        cache_type = data['type']
        if cache_type == CacheType.Filesystem and not data.get('path'):
            raise ConfigurationError({'path': 'cache.path is required for a filesystem cache'})
        if cache_type in (CacheType.Redis, CacheType.Memcached) and not data.get('url'):
            raise ConfigurationError({'url': f'cache.url is required for a {cache_type.name} cache'})
        return data

    parser.add_option('cache', default={}, type=cache_validator)

    s3_parser = ConfigParser()
    s3_parser.add_option('access_key_id', type=str, default=None)
    s3_parser.add_option('secret_access_key', type=str, default=None)
    s3_parser.add_option('endpoint', type=str, default=None)
    s3_parser.add_option('region', type=str, default=None)
    s3_parser.add_option('thumbnails_bucket', type=str, default='thumbnails')
    s3_parser.add_option('staging_bucket', type=str, default='staging')
    s3_parser.add_option('ponos_logs_bucket', type=str, default='ponos-logs')
    s3_parser.add_option('ponos_artifacts_bucket', type=str, default='ponos-artifacts')

    def s3_validator(value):
        data = s3_parser.parse_data(value)
        if not data.get('access_key_id') and not data.get('secret_access_key'):
            # No configuration specified;
            # just ignore and let the system checks warn about this without preventing startup
            return data
        if not data.get('endpoint') and not data.get('region'):
            raise ConfigurationError('One of `s3.endpoint` or `s3.region` are required')
        return data

    parser.add_option('s3', type=s3_validator, default={})

    return parser
