from django.contrib.auth.models import AnonymousUser
from django.test import TestCase

from arkindex.documents.models import Corpus
from arkindex.users.models import Role, User


class TestACL(TestCase):
    '''
    Test Corpus manager ACL helpers
    '''

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.anon = AnonymousUser()
        cls.user = User.objects.create_user('user@address.com', 'P4$5w0Rd')
        cls.admin = User.objects.create_superuser('admin@address.com', 'P4$5w0Rd')
        cls.internal = User.objects.create_user('internal@address.com', 'P4$5w0Rd', internal=True)

        cls.corpus_public = Corpus.objects.create(name='A Public', public=True)
        cls.corpus_private = Corpus.objects.create(name='B Private')
        cls.corpus_private.memberships.create(user=cls.user, level=Role.Guest.value)
        cls.corpus_hidden = Corpus.objects.create(name='C Hidden')

    def test_anon(self):
        # An anonymous user has only access to public
        self.assertCountEqual(
            Corpus.objects.readable(self.anon).values_list('id', flat=True),
            [self.corpus_public.id]
        )

    def test_user(self):
        # An anonymous user has access to public & private
        self.assertFalse(self.user.is_admin)
        self.assertCountEqual(
            Corpus.objects.readable(self.user).values_list('id', flat=True),
            [
                self.corpus_private.id,
                *Corpus.objects.filter(public=True).values_list('id', flat=True)
            ]
        )

    def test_admin(self):
        # An admin has access with admin privileges to all corpora
        self.assertTrue(self.admin.is_admin)
        self.assertCountEqual(
            Corpus.objects.admin(self.admin).values_list('id'),
            Corpus.objects.all().values_list('id')
        )

    def test_internal(self):
        # Internal users have access to all corpora
        self.assertTrue(self.internal.is_internal)
        self.assertCountEqual(
            Corpus.objects.admin(self.internal).values_list('id'),
            Corpus.objects.all().values_list('id')
        )
